import com.jniwrapper.*;

/**
 * @author Alexander Kireyev
 */
public class Buzzer
{
    private Library _kernel;
    private Library _user;

    private static final int TIMER_ID = 1;
    private static final String WINDOW_CLASS_NAME = "BuzzerWindowClass";
    private Pointer.Void _hSplash;
    private Pointer.Void _hSplashImage;
    private Library _gdi;
    private static final int SPLASH_WIDTH = 122;

    public Buzzer()
    {
        _kernel = new Library("kernel32");
        _user = new Library("user32");
        _gdi = new Library("gdi32");
        Int brType = new Int(0); // white brush
        Pointer.Void bgBrush = new Pointer.Void();
        _gdi.getFunction("GetStockObject").invoke(bgBrush, brType);
        if (bgBrush.getValue() == 0)
        {
            throw new RuntimeException("Failed to get background brush, error code = " + getLastError(true));
        }
        WndClass wndClass = new WndClass(new WindowProc(), WINDOW_CLASS_NAME, bgBrush);
        Int16 result = new Int16();
        _user.getFunction("RegisterClassA").invoke(result, new Pointer(wndClass));
        if (result.getValue() == 0)
        {
            throw new RuntimeException("Failed to register window class, error code = " + getLastError(true));
        }
        _hSplash = new Pointer.Void();
        Pointer.Void hModule = new Pointer.Void();
        _kernel.getFunction("GetModuleHandleA").invoke(hModule, new Pointer.Void(0));
        // There is no CreateWindowA(W) function - they are #defined throug CreateWindowEx with first argument
        // equal to 0.
        _user.getFunction("CreateWindowExA").invoke(_hSplash,
                new Parameter[]
                {
                    new UInt32(0),
                    new AnsiString(WINDOW_CLASS_NAME),
                    new AnsiString("Buzzer"),
                    new UInt32(0x80000000),         // WS_POPUP
                    new Int(0),
                    new Int(0),
                    new Int(0),
                    new Int(0),
                    new Pointer.Void(0),
                    new Pointer.Void(0),
                    hModule,
                    new Pointer.Void(0),

                });
        if (_hSplash.getValue() == 0)
        {
            throw new RuntimeException("Failed to create splash window, error code = " + getLastError(true));
        }
    }

    public static void main(String[] args)
    {
        // Add JNIWrap.dll directory to library loader search path
        DefaultLibraryLoader.getInstance().addPath("Bin");
        Buzzer buzzer = new Buzzer();
        if (!buzzer.checkOneInstance())
        {
            buzzer.error("Buzzer is already running");
            System.exit(0);
        }
        buzzer.startTimer(10000);
        buzzer.run();
    }

    private void run()
    {
        Pointer.Void hDesktop = new Pointer.Void();
        _user.getFunction("GetDesktopWindow").invoke(hDesktop);
        Int width = new Int();
        Int height = new Int();
        _user.getFunction("GetWindowRect").invoke(null,
                hDesktop,
                new Pointer.OutOnly(new Structure(new Parameter[] {new Int(), new Int(), width, height})));
        Pointer.Void hRgn = new Pointer.Void();
        _gdi.getFunction("CreateEllipticRgn").invoke(hRgn,
                new Int(0), new Int(0),
                new Int(SPLASH_WIDTH), new Int(SPLASH_WIDTH));
        _user.getFunction("SetWindowRgn").invoke(null, _hSplash, hRgn, new Bool(true));
        Bool result = new Bool();
        _user.getFunction("SetWindowPos").invoke(result,
                new Parameter[]
                {
                    _hSplash,
                    new Pointer.Void(-1),   // HWND_TOPMOST
                    new Int((width.getValue() - SPLASH_WIDTH) / 2),
                    new Int((height.getValue() - SPLASH_WIDTH) / 2),
                    new Int(SPLASH_WIDTH),
                    new Int(SPLASH_WIDTH),
                    new UInt(0x40)          // SWP_SHOWWINDOW
                });
        if (!result.getValue())
        {
            throw new RuntimeException("Failed to show splash, error code = " + getLastError(true));
        }
        Function getMessage = _user.getFunction("GetMessageA", null);
        Function translateMessage = _user.getFunction("TranslateMessage", null);
        Function dispatchMessage = _user.getFunction("DispatchMessageA", null);
        MSG msg = new MSG();
        Pointer msgPointer = new Pointer(msg);
        for (;;)
        {
            getMessage.invoke(result, msgPointer, _hSplash, new UInt32(0), new UInt32(0));
            if (!result.getValue())
            {
                break;
            }
            translateMessage.invoke(null, msgPointer);
            dispatchMessage.invoke(null, msgPointer);
        }

    }

    private void paintSplash()
    {
        Pointer.Void hDC = new Pointer.Void();
        PrimitiveArray paintStruct = new PrimitiveArray(new Int8(), 64);
        _user.getFunction("BeginPaint").invoke(hDC, _hSplash, new Pointer.OutOnly(paintStruct));
        if (hDC.isNull())
        {
            throw new RuntimeException("Failed to get DC, error code = " + getLastError(true));
        }
        Pointer.Void splashDC = new Pointer.Void();
        _gdi.getFunction("CreateCompatibleDC").invoke(splashDC, hDC);
        _gdi.getFunction("SelectObject").invoke(null, splashDC, getSplashImage());
        _gdi.getFunction("BitBlt").invoke(null,
                new Parameter[]
                {
                    hDC,
                    new Int(0),
                    new Int(0),
                    new Int(SPLASH_WIDTH),
                    new Int(SPLASH_WIDTH),
                    splashDC,
                    new Int(0),
                    new Int(0),
                    new UInt32(0x00CC0020) // SRCCOPY
                });
        _gdi.getFunction("DeleteDC").invoke(null, splashDC);
        _user.getFunction("EndPaint").invoke(null, _hSplash, new Pointer.Const(paintStruct));
    }

    private Pointer.Void getSplashImage()
    {
        if (_hSplashImage == null)
        {
            _hSplashImage = new Pointer.Void();
            _user.getFunction("LoadImageA").invoke(
                    _hSplashImage,
                    new Parameter[]
                    {
                        new Pointer.Void(0),
                        new AnsiString("Buzzer.bmp"),
                        new UInt(0),
                        new Int(0),
                        new Int(0),
                        new UInt(0x8010) // LR_LOADFROMFILE | LR_SHARED
                    });
            if (_hSplashImage.isNull())
            {
                throw new RuntimeException("Failed to load image, error code = " + getLastError(true));
            }
        }
        Pointer.Void hSplashImage = _hSplashImage;
        return hSplashImage;
    }

    private void startTimer(long timeout)
    {
        UInt eventID = new UInt(TIMER_ID);
        UInt timeOutVal = new UInt(timeout);
        UInt result = new UInt();
        TimeOutCallback timeOutCallback = new TimeOutCallback();
        Function setTimer = _user.getFunction("SetTimer");
        setTimer.invoke(result, _hSplash, eventID, timeOutVal, timeOutCallback);
        if (result.getValue() == 0)
        {
            throw new RuntimeException("Failed to create a timer, error code = " + getLastError(true));
        }
    }

    private void stopTimer(Pointer.Void hwnd, UInt timerID)
    {
        Function killTimer = _user.getFunction("KillTimer");
        killTimer.invoke(null, hwnd, timerID);
    }

    private void error(String message)
    {
        showMessageBox(message, 0x10);
    }

    private void showMessageBox(String message, int flags)
    {
        Pointer.Void hWnd = new Pointer.Void(0);
        AnsiString text = new AnsiString(message);
        AnsiString caption = new AnsiString("Buzzer");
        UInt uFlags = new UInt(flags);
        // For functions taking strings as parameters there are two versions: ANSI and Unicode
	// We chose ANSI here and hence fucncion name ends with A
	// It should be noted that there is no plain MessageBox function exported
        _user.getFunction("MessageBoxA").invoke(null, hWnd, text, caption, uFlags);
    }

    private void message(String message)
    {
        showMessageBox(message, 0x30);
    }

    private boolean checkOneInstance()
    {
        AnsiString mutexName = new AnsiString("com.jniwrapper.sample.BuzzerMutex");
        UInt32 desiredAcces = new UInt32(0x1F0001);
        Bool inheritHandle = new Bool(false);
        Pointer.Void mutexHandle = new Pointer.Void();
        Function openMutex = _kernel.getFunction("OpenMutexA");
        openMutex.invoke(mutexHandle, desiredAcces, inheritHandle, mutexName);
        if (!mutexHandle.isNull())
        {
            // Mutex exists - one instance is already running
            return false;
        }

        // Not yet running - lock by creating mutex
        Pointer.Void mutexAttributes = new Pointer.Void(0);
        Bool initialOwner = new Bool(false);
        Function createMutex = _kernel.getFunction("CreateMutexA");
        createMutex.invoke(mutexHandle, mutexAttributes, initialOwner, mutexName);
        if (mutexHandle.isNull())
        {
            throw new RuntimeException("Mutex creation failed, last error = " + getLastError(true));
        }
        return true;
    }

    private long getLastError(boolean clear)
    {
        UInt32 r = new UInt32();
        Function getLE = _kernel.getFunction("GetLastError");
        getLE.invoke(r);
        long errCode = r.getValue();
        if (clear)
        {
            UInt32 zero = new UInt32(0);
            Function setLE = _kernel.getFunction("SetLastError");
            setLE.invoke(null, zero);
        }
        return errCode;
    }

    private void buzz()
    {
        // Prepare beep parameters: low and high frequencies and beep durations
        UInt32 low = new UInt32(400);
        UInt32 high = new UInt32(1000);
        UInt32 durLow = new UInt32(200);
        UInt32 durHigh = new UInt32(200);

        // Obtain function reference from the library
        Function beep = _kernel.getFunction("Beep");

        // Do beeping
        for (int i = 0; i < 5; i++)
        {
            beep.invoke(null, high, durHigh);
            beep.invoke(null, low, durLow);
        }
    }
    private void hideSplash()
    {
        _user.getFunction("ShowWindow").invoke(null, _hSplash, new Int(0));
    }

    private class TimeOutCallback extends Callback
    {
        private Pointer.Void _hwnd = new Pointer.Void();
        private UInt _msg = new UInt();
        private UInt _timerID = new UInt();
        private UInt32 _time = new UInt32();

        public TimeOutCallback()
        {
            init(new Parameter[] {
                _hwnd,
                _msg,
                _timerID,
                _time
            }, null);
        }

        public void callback()
        {
            stopTimer(_hwnd, _timerID);
            buzz();
            hideSplash();
            message("Timer has elapsed!");
            System.exit(0);
        }
    }

    private static class WndClass extends Structure
    {
        private UInt32 _style = new UInt32();
        private Callback _lpfnWndProc;
        private Int32 _cbClsExtra = new Int32();
        private Int32 _cbWndExtra = new Int32();
        private Pointer.Void _hInstance = new Pointer.Void();
        private Pointer.Void _hIcon = new Pointer.Void();
        private Pointer.Void _hCursor = new Pointer.Void();
        private Pointer.Void _hbrBackground = new Pointer.Void();
        private AnsiString _lpszClassName = new AnsiString();

        public WndClass(Callback windowProc, String className, Pointer.Void bgBrushHandle)
        {
            _style.setValue(3); // CS_HREDRAW | CS_VREDRAW
            _lpfnWndProc = windowProc;
            _lpszClassName.setValue(className);
            _cbClsExtra.setValue(0);
            _cbWndExtra.setValue(0);
            _hInstance.setValue(0);
            _hIcon.setValue(0);
            _hCursor.setValue(0);
            _hbrBackground.setValue(bgBrushHandle.getValue());
            init(new Parameter[] {_style, _lpfnWndProc, _cbClsExtra, _cbWndExtra, _hInstance, _hIcon, _hCursor, _hbrBackground, new Pointer.Void(0), new Pointer(_lpszClassName)});
        }
    }

    private static class POINT extends Structure
    {
        public LongInt _x = new LongInt();
        public LongInt _y = new LongInt();

        public POINT()
        {
            init(new Parameter[] {_x, _y});
        }
    }

    private static class MSG extends Structure
    {
        Pointer.Void _hwnd = new Pointer.Void();
        UInt _message = new UInt();
        UInt _wParam = new UInt();
        UInt _lParam = new UInt();
        Int  _time = new Int();
        POINT _pt = new POINT();

        public MSG()
        {
            init(new Parameter[]{_hwnd, _message, _wParam, _lParam, _time, _pt});
        }
    }

    private class WindowProc extends Callback
    {
        Int32 _hwnd = new Int32();
        UInt32 _msg = new UInt32();
        Int32 _wParam = new Int32();
        Int32 _lParam = new Int32();
        UInt32 _lResult = new UInt32();
        private Function _defWindowProc;

        public WindowProc()
        {
            init(new Parameter[] {_hwnd, _msg, _wParam, _lParam}, _lResult);
            Library user32 = new Library("user32");
            _defWindowProc = user32.getFunction("DefWindowProcA", null);
        }

        public void callback()
        {
            switch ((int)_msg.getValue())
            {
                case 0xF:   // WM_PAINT
                    paintSplash();
                default:
                    _defWindowProc.invoke(_lResult, _hwnd, _msg, _wParam, _lParam);
                    break;
            }
        }
    }
}
